//
// I2CLiquidCrystal.h
//
// Class to control I2C LCD
// 
// Written by Nobuki HIRAMINE
// http://www.hiramine.com
// Last modified : 2017/07/02
//

#include <Wire.h>

class I2CLiquidCrystal : public Print
{
private:
	int _numlines = 2;
	int m_i2caddr_lcd = 0;
	int m_pin_sda = 0;
	int m_pin_scl = 0;

public:
	I2CLiquidCrystal( int i2caddr_lcd, int pin_sda, int pin_scl )
	{
		m_i2caddr_lcd = i2caddr_lcd;
		m_pin_sda = pin_sda;
		m_pin_scl = pin_scl;
	}

	void begin()
	{
		// Init I2C
		Wire.begin(m_pin_sda, m_pin_scl);

		// Function Set
		// 2進数表現で、001L NF00 形式のデータを送信する。
		// L : データ長。1 = 8ビット, 0 = 4ビット。AQM1602XAは8bitなので1。
		// N : Line数。1 = 2行。0 = 1行。AQM1602XAは2行なので1。
		// F : フォントタイプ。1 = 5x10ドット, 0 = 5x8ドット。AQM1602XAは5x8ドットなので0。
		// 以上により、2進数表現で、0011 1000。16進数表現で、0x38のデータを送信する。
		_command( 0x38 );

		// 拡張コマンドモードに移行（内部発振周波数、コントラスト等を設定するため）
		// Function Setで送ったデータの1ビット目を立てたデータを送る
		// 0x38を送った場合は、1足して、0x39を送る
		_command( 0x39 );

		// 内部発振周波数
		// 2進数表現で、0001 [BS][F2][F1][F0] 形式のデータを送信する。
		// データシートより、2進数表現で、0001 0100。16進数表現で、0x14のデータを送信する。
		_command( 0x14 );

		// コントラスト（1/2）
		// 2進数表現で、0111 [C3][C2][C1][C0] 形式のデータを送信する。
		// データシートより、「5Vの場合、C5=1,C4=0,C3=0,C2=0,C1=1,C0=1で少し濃い目ぐらい」
		// 3.3Vで使用するので「C5=1,C4=0,C3=0,C2=0,C1=0,C0=0」とする。
		// 2進数表現で、0111 0000。16進数表現で、0x70のデータを送信する。
		_command( 0x70 );

		// コントラスト（2/2）
		// 2進数表現で、0101 [Ion][Bon][C5][C4] 形式のデータを送信する。
		// Ion : ICON display on/off。1 = on, 0 = off。Offとしたいので0。
		// Bon : Booster circuit on/off。3.3Vの場合はBooster回路をOnとするので1。
		// コントラスト（1/2）より「C5=1,C4=0,C3=0,C2=0,C1=0,C0=0」。
		// 2進数表現で、0101 0110。16進数表現で、0x56のデータを送信する。
		_command( 0x56 );

		// Follower control
		// 2進数表現で、0110 [Fon][Rab2][Rab1][Rab0] 形式のデータを送信する。
		// データシートより、2進数表現で、0110 1100。16進数表現で、0x6Cのデータを送信する。
		_command( 0x6C );

		// 200msec : for power stable
		delay(200);

		// 拡張コマンドモードの終了
		// Function Setで送ったデータの1ビット目を倒したデータを送る
		_command( 0x38 );

		// Display ON/OFF Control
		// 2進数表現で、0000 1DCB 形式のデータを送信する。
		// D : Display On/Off。1 = On, 0 = Off。Display On なので1。
		// C : Cursor On/Off。1 = On, 0 = Off。Cursor Offとしたいので0。
		// B : Blinking of cursor On/Off。1 = On, 0 = Off。Blinking Offとしたいので0。
		// 以上により、2進数表現で、0000 1100。16進数表現で、0x0Cのデータを送信する。
		_command(0x0C);

		// Entry Mode Set
		// 2進数表現で、0000 01DS 形式のデータを送信する。
		// D : Cursor move direction。1 = Increment, 0 = Decrement。カーソル移動方向は増加方向なので1。
		// S : Shift the display。1 = Shiftする, 0 = Shiftしない。Shiftしないとしたいので0。
		// 以上により、2進数表現で、0000 0110。16進数表現で、0x06のデータを送信する。
		_command(0x06);

		clear();		// Clear display
	}

	void clear()
	{
		_command( 0x01 );
		delayMicroseconds(1060);	// need 1.06msec
	}

	void setCursor( uint8_t col, uint8_t row )
	{
		static const uint8_t row_offsets[2] = {0x00, 0x40};
		if( row >= _numlines )
		{
			row = _numlines - 1;
		}
		_command( 0x80 | ( row_offsets[row] + col ) );
	}

	size_t write(uint8_t value)
	{
		_send(value, HIGH);
		return 1;
	}

private:
	void _command( uint8_t value )
	{
		_send( value, LOW );
		delayMicroseconds(27);	// need 26.3microsec
	}

	void _send( uint8_t value, uint8_t mode )
	{
	    Wire.beginTransmission(m_i2caddr_lcd);
		if( mode )
		{	// print mode
		    Wire.write(0x40);
		}
		else
		{	// command mode
		    Wire.write(0x00);
		}
	    Wire.write(value);
	    Wire.endTransmission();
	}
};
